"""Module for handling network connections to Roborock devices.

This is used internally by the device manager for creating connections to
Roborock devices. These modules contain common code, not specific to a
particular device or application level protocol.
"""

__all__: list[str] = []
